@@ -41,6 +41,8 @@
 #include "gltexture_p.h"
 
 #include <private/qdebug_p.h>
+#include <private/qopengltexture_p.h>
+#include <private/qopengltexturehelper_p.h>
 #include <QDebug>
 #include <QOpenGLFunctions>
 #include <QOpenGLTexture>
@@ -49,13 +51,9 @@
 #include <Qt3DRender/qtexturedata.h>
 #include <Qt3DRender/qtextureimagedata.h>
 #include <Qt3DRender/private/managers_p.h>
-#include <Qt3DRender/private/texturedatamanager_p.h>
 #include <Qt3DRender/private/qabstracttexture_p.h>
 #include <Qt3DRender/private/renderbuffer_p.h>
 #include <Qt3DRender/private/qtextureimagedata_p.h>
-#include <Qt3DCore/qpropertyupdatedchange.h>
-#include <Qt3DCore/qpropertynodeaddedchange.h>
-#include <Qt3DCore/qpropertynoderemovedchange.h>
 
 #if !defined(QT_OPENGL_ES_2)
 #include <QOpenGLFunctions_3_1>
@@ -69,59 +67,110 @@ using namespace Qt3DCore;
 namespace Qt3DRender {
 namespace Render {
 
-GLTexture::GLTexture(TextureDataManager *texDataMgr,
-                     TextureImageDataManager *texImgDataMgr,
-                     const QTextureGeneratorPtr &texGen,
-                     bool unique)
-    : m_unique(unique)
+namespace {
+
+// This uploadGLData where the data is a fullsize subimage
+// as QOpenGLTexture doesn't allow partial subimage uploads
+void uploadGLData(QOpenGLTexture *glTex,
+                  int level, int layer, QOpenGLTexture::CubeMapFace face,
+                  const QByteArray &bytes, const QTextureImageDataPtr &data)
+{
+    if (data->isCompressed()) {
+        glTex->setCompressedData(level, layer, face, bytes.size(), bytes.constData());
+    } else {
+        QOpenGLPixelTransferOptions uploadOptions;
+        uploadOptions.setAlignment(1);
+        glTex->setData(level, layer, face, data->pixelFormat(), data->pixelType(), bytes.constData(), &uploadOptions);
+    }
+}
+
+// For partial sub image uploads
+void uploadGLData(QOpenGLTexture *glTex,
+                  int mipLevel, int layer, QOpenGLTexture::CubeMapFace cubeFace,
+                  int xOffset, int yOffset, int zOffset,
+                  const QByteArray &bytes, const QTextureImageDataPtr &data)
+{
+    if (data->isCompressed()) {
+        qWarning() << Q_FUNC_INFO << "Uploading non full sized Compressed Data not supported yet";
+    } else {
+        QOpenGLPixelTransferOptions uploadOptions;
+        uploadOptions.setAlignment(1);
+        glTex->setData(xOffset, yOffset, zOffset,
+                       data->width(), data->height(), data->depth(),
+                       mipLevel, layer, cubeFace, data->layers(),
+                       data->pixelFormat(), data->pixelType(),
+                       bytes.constData(), &uploadOptions);
+    }
+}
+
+} // anonymous
+
+
+GLTexture::GLTexture()
+    : m_dirtyFlags(None)
     , m_gl(nullptr)
     , m_renderBuffer(nullptr)
-    , m_textureDataManager(texDataMgr)
-    , m_textureImageDataManager(texImgDataMgr)
-    , m_dataFunctor(texGen)
+    , m_dataFunctor()
     , m_pendingDataFunctor(nullptr)
     , m_sharedTextureId(-1)
     , m_externalRendering(false)
+    , m_wasTextureRecreated(false)
 {
-    // make sure texture generator is executed
-    // this is needed when Texture have the TargetAutomatic
-    // to ensure they are loaded before trying to instantiate the QOpenGLTexture
-    if (!texGen.isNull())
-        m_textureDataManager->requestData(texGen, this);
 }
 
 GLTexture::~GLTexture()
 {
-    destroyGLTexture();
 }
 
-void GLTexture::destroyResources()
-{
-    if (m_dataFunctor)
-        m_textureDataManager->releaseData(m_dataFunctor, this);
-}
-
-void GLTexture::destroyGLTexture()
+// Must be called from RenderThread with active GL context
+void GLTexture::destroy()
 {
     delete m_gl;
     m_gl = nullptr;
     delete m_renderBuffer;
     m_renderBuffer = nullptr;
 
-    m_dirtyFlags.store(0);
+    m_dirtyFlags = None;
+    m_sharedTextureId = -1;
+    m_externalRendering = false;
+    m_wasTextureRecreated = false;
+    m_dataFunctor.reset();
+    m_pendingDataFunctor = nullptr;
 
-    destroyResources();
+    m_properties = {};
+    m_parameters = {};
+    m_textureData.reset();
+    m_images.clear();
+    m_imageData.clear();
+    m_pendingTextureDataUpdates.clear();
 }
 
 bool GLTexture::loadTextureDataFromGenerator()
 {
-    m_textureData = m_textureDataManager->getData(m_dataFunctor);
+    m_textureData = m_dataFunctor->operator()();
     // if there is a texture generator, most properties will be defined by it
     if (m_textureData) {
-        if (m_properties.target != QAbstractTexture::TargetAutomatic)
-            qWarning() << "[Qt3DRender::GLTexture] When a texture provides a generator, it's target is expected to be TargetAutomatic";
+        const QAbstractTexture::Target target = m_textureData->target();
 
-        m_actualTarget = m_textureData->target();
+        // If both target and functor return Automatic we are still
+        // probably loading the texture, return false
+        if (m_properties.target == QAbstractTexture::TargetAutomatic &&
+            target == QAbstractTexture::TargetAutomatic) {
+            m_textureData.reset();
+            return false;
+        }
+
+        if (m_properties.target != QAbstractTexture::TargetAutomatic &&
+            target != QAbstractTexture::TargetAutomatic &&
+            m_properties.target != target) {
+            qWarning() << Q_FUNC_INFO << "Generator and Properties not requesting the same texture target";
+            m_textureData.reset();
+            return false;
+        }
+
+        // We take target type from generator if it wasn't explicitly set by the user
+        if (m_properties.target == QAbstractTexture::TargetAutomatic)
+            m_properties.target = target;
         m_properties.width = m_textureData->width();
         m_properties.height = m_textureData->height();
         m_properties.depth = m_textureData->depth();
@@ -143,7 +192,7 @@ void GLTexture::loadTextureDataFromImages()
 {
     int maxMipLevel = 0;
     for (const Image &img : qAsConst(m_images)) {
-        const QTextureImageDataPtr imgData = m_textureImageDataManager->getData(img.generator);
+        const QTextureImageDataPtr imgData = img.generator->operator()();
         // imgData may be null in the following cases:
         // - Texture is created with TextureImages which have yet to be
         // loaded (skybox where you don't yet know the path, source set by
@@ -179,26 +228,23 @@ void GLTexture::loadTextureDataFromImages()
     }
 }
 
+// Called from RenderThread
 GLTexture::TextureUpdateInfo GLTexture::createOrUpdateGLTexture()
 {
-    QMutexLocker locker(&m_textureMutex);
-    bool needUpload = false;
     TextureUpdateInfo textureInfo;
-
-    m_properties.status = QAbstractTexture::Error;
+    m_wasTextureRecreated = false;
 
     const bool hasSharedTextureId = m_sharedTextureId > 0;
-
     // Only load texture data if we are not using a sharedTextureId
+    // Check if dataFunctor or images have changed
     if (!hasSharedTextureId) {
-        // on the first invocation in the render thread, make sure to
-        // evaluate the texture data generator output
-        // (this might change some property values)
-        if (m_dataFunctor && !m_textureData) {
+        // If dataFunctor exists and we have no data and it hasn´t run yet
+        if (m_dataFunctor && !m_textureData && m_dataFunctor.get() != m_pendingDataFunctor ) {
             const bool successfullyLoadedTextureData = loadTextureDataFromGenerator();
+            // If successful, m_textureData has content
             if (successfullyLoadedTextureData) {
                 setDirtyFlag(Properties, true);
-                needUpload = true;
+                setDirtyFlag(TextureData, true);
             } else {
                 if (m_pendingDataFunctor != m_dataFunctor.get()) {
                     qWarning() << "[Qt3DRender::GLTexture] No QTextureData generated from Texture Generator yet. Texture will be invalid for this frame";
@@ -209,56 +255,77 @@ GLTexture::TextureUpdateInfo GLTexture::createOrUpdateGLTexture()
             }
         }
 
-        // additional texture images may be defined through image data generators
-        if (testDirtyFlag(TextureData)) {
+        // If images have changed, clear previous images data
+        // and regenerate m_imageData for the images
+        if (testDirtyFlag(TextureImageData)) {
             m_imageData.clear();
             loadTextureDataFromImages();
-            needUpload = true;
+            // Mark for upload if we actually have something to upload
+            if (!m_imageData.empty()) {
+                setDirtyFlag(TextureData, true);
+            }
+            // Reset image flag
+            setDirtyFlag(TextureImageData, false);
         }
 
-        // don't try to create the texture if the format was not set
-        if (m_properties.format == QAbstractTexture::Automatic) {
+        // Don't try to create the texture if the target or format was still not set
+        // Format should either be set by user or if Automatic
+        // by either the dataGenerator of the texture or the first Image
+        // Target should explicitly be set by the user or the dataGenerator
+        if (m_properties.target == QAbstractTexture::TargetAutomatic ||
+            m_properties.format == QAbstractTexture::Automatic ||
+            m_properties.format == QAbstractTexture::NoFormat) {
             textureInfo.properties.status = QAbstractTexture::Error;
             return textureInfo;
         }
     }
 
-    // if the properties changed, we need to re-allocate the texture
+    // If the properties changed or texture has become a shared texture from a
+    // 3rd party engine, we need to destroy and maybe re-allocate the texture
     if (testDirtyFlag(Properties) || testDirtyFlag(SharedTextureId)) {
         delete m_gl;
         m_gl = nullptr;
         textureInfo.wasUpdated = true;
-        // If we are destroyed because of some property change but still our content data
-        // make sure we are marked for upload
-        if (m_textureData || !m_imageData.empty())
-            needUpload = true;
+        // If we are destroyed because of some property change but still have (some) of
+        // our content data make sure we are marked for upload
+        // TO DO: We should actually check if the textureData is still correct
+        // in regard to the size, target and format of the texture though.
+        if (!testDirtyFlag(SharedTextureId) &&
+            (m_textureData || !m_imageData.empty() || !m_pendingTextureDataUpdates.empty()))
+            setDirtyFlag(TextureData, true);
     }
 
     m_properties.status = QAbstractTexture::Ready;
 
-    if (hasSharedTextureId && testDirtyFlag(SharedTextureId)) {
+    if (testDirtyFlag(SharedTextureId) || hasSharedTextureId) {
         // Update m_properties by doing introspection on the texture
-        introspectPropertiesFromSharedTextureId();
+        if (hasSharedTextureId)
+            introspectPropertiesFromSharedTextureId();
+        setDirtyFlag(SharedTextureId, false);
     } else {
         // We only build a QOpenGLTexture if we have no shared textureId set
         if (!m_gl) {
             m_gl = buildGLTexture();
             if (!m_gl) {
+                qWarning() << "[Qt3DRender::GLTexture] failed to create texture";
                 textureInfo.properties.status = QAbstractTexture::Error;
                 return textureInfo;
             }
 
             m_gl->allocateStorage();
             if (!m_gl->isStorageAllocated()) {
+                qWarning() << "[Qt3DRender::GLTexture] failed to allocate texture";
                 textureInfo.properties.status = QAbstractTexture::Error;
                 return textureInfo;
             }
+            m_wasTextureRecreated = true;
         }
 
         textureInfo.texture = m_gl;
 
         // need to (re-)upload texture data?
-        if (needUpload) {
+        const bool needsUpload = testDirtyFlag(TextureData);
+        if (needsUpload) {
             uploadGLTextureData();
             setDirtyFlag(TextureData, false);
         }
@@ -266,26 +333,20 @@ GLTexture::TextureUpdateInfo GLTexture::createOrUpdateGLTexture()
         // need to set texture parameters?
         if (testDirtyFlag(Properties) || testDirtyFlag(Parameters)) {
             updateGLTextureParameters();
+            setDirtyFlag(Properties, false);
+            setDirtyFlag(Parameters, false);
         }
     }
 
     textureInfo.properties = m_properties;
 
-    // un-set properties and parameters. The TextureData flag might have been set by another thread
-    // in the meantime, so don't clear that.
-    setDirtyFlag(Properties, false);
-    setDirtyFlag(Parameters, false);
-    setDirtyFlag(SharedTextureId, false);
-
     return textureInfo;
 }
 
 RenderBuffer *GLTexture::getOrCreateRenderBuffer()
 {
-    QMutexLocker locker(&m_textureMutex);
-
     if (m_dataFunctor && !m_textureData) {
-        m_textureData = m_textureDataManager->getData(m_dataFunctor);
+        m_textureData = m_dataFunctor->operator()();
         if (m_textureData) {
             if (m_properties.target != QAbstractTexture::TargetAutomatic)
                 qWarning() << "[Qt3DRender::GLTexture] [renderbuffer] When a texture provides a generator, it's target is expected to be TargetAutomatic";
@@ -318,9 +379,15 @@ RenderBuffer *GLTexture::getOrCreateRenderBuffer()
     return m_renderBuffer;
 }
 
+// This must be called from the RenderThread
+// So GLTexture release from the manager can only be done from that thread
+void GLTexture::cleanup()
+{
+    destroy();
+}
+
 void GLTexture::setParameters(const TextureParameters &params)
 {
-    QMutexLocker locker(&m_textureMutex);
     if (m_parameters != params) {
         m_parameters = params;
         setDirtyFlag(Parameters);
@@ -329,10 +396,8 @@ void GLTexture::setParameters(const TextureParameters &params)
 
 void GLTexture::setProperties(const TextureProperties &props)
 {
-    QMutexLocker locker(&m_textureMutex);
     if (m_properties != props) {
         m_properties = props;
-        m_actualTarget = props.target;
         setDirtyFlag(Properties);
     }
 }
@@ -353,25 +418,16 @@ void GLTexture::setImages(const QVector<Image> &images)
 
     if (!same) {
         m_images = images;
-        requestUpload();
+        requestImageUpload();
     }
 }
 
 void GLTexture::setGenerator(const QTextureGeneratorPtr &generator)
 {
-    // Note: we do not compare if the generator is different
-    // as in some cases we may want to reset the same generator to force a reload
-    // e.g when using remote urls for textures
-    if (m_dataFunctor)
-        m_textureDataManager->releaseData(m_dataFunctor, this);
-
     m_textureData.reset();
     m_dataFunctor = generator;
-
-    if (m_dataFunctor) {
-        m_textureDataManager->requestData(m_dataFunctor, this);
-        requestUpload();
-    }
+    m_pendingDataFunctor = nullptr;
+    requestUpload();
 }
 
 void GLTexture::setSharedTextureId(int textureId)
@@ -382,6 +438,12 @@ void GLTexture::setSharedTextureId(int textureId)
     }
 }
 
+void GLTexture::addTextureDataUpdates(const QVector<QTextureDataUpdate> &updates)
+{
+    m_pendingTextureDataUpdates += updates;
+    requestUpload();
+}
+
 // Return nullptr if
 // - context cannot be obtained
 // - texture hasn't yet been loaded
@@ -393,28 +455,29 @@ QOpenGLTexture *GLTexture::buildGLTexture()
         return nullptr;
     }
 
-    if (m_actualTarget == QAbstractTexture::TargetAutomatic) {
+    const QAbstractTexture::Target actualTarget = m_properties.target;
+    if (actualTarget == QAbstractTexture::TargetAutomatic) {
         // If the target is automatic at this point, it means that the texture
         // hasn't been loaded yet (case of remote urls) and that loading failed
         // and that target format couldn't be deduced
         return nullptr;
     }
 
-    QOpenGLTexture* glTex = new QOpenGLTexture(static_cast<QOpenGLTexture::Target>(m_actualTarget));
+    QOpenGLTexture* glTex = new QOpenGLTexture(static_cast<QOpenGLTexture::Target>(actualTarget));
 
     // m_format may not be ES2 compatible. Now it's time to convert it, if necessary.
     QAbstractTexture::TextureFormat format = m_properties.format;
     if (ctx->isOpenGLES() && ctx->format().majorVersion() < 3) {
         switch (m_properties.format) {
-        case QOpenGLTexture::RGBA8_UNorm:
-        case QOpenGLTexture::RGBAFormat:
+        case QAbstractTexture::RGBA8_UNorm:
+        case QAbstractTexture::RGBAFormat:
             format = QAbstractTexture::RGBAFormat;
             break;
-        case QOpenGLTexture::RGB8_UNorm:
-        case QOpenGLTexture::RGBFormat:
+        case QAbstractTexture::RGB8_UNorm:
+        case QAbstractTexture::RGBFormat:
             format = QAbstractTexture::RGBFormat;
             break;
-        case QOpenGLTexture::DepthFormat:
+        case QAbstractTexture::DepthFormat:
             format = QAbstractTexture::DepthFormat;
             break;
         default:
@@ -436,19 +499,19 @@ QOpenGLTexture *GLTexture::buildGLTexture()
     }
 
     glTex->setFormat(m_properties.format == QAbstractTexture::Automatic ?
-                         QOpenGLTexture::NoFormat :
-                         static_cast<QOpenGLTexture::TextureFormat>(format));
+                     QOpenGLTexture::NoFormat :
+                     static_cast<QOpenGLTexture::TextureFormat>(format));
     glTex->setSize(m_properties.width, m_properties.height, m_properties.depth);
     // Set layers count if texture array
-    if (m_actualTarget == QAbstractTexture::Target1DArray ||
-        m_actualTarget == QAbstractTexture::Target2DArray ||
-        m_actualTarget == QAbstractTexture::Target2DMultisampleArray ||
-        m_actualTarget == QAbstractTexture::TargetCubeMapArray) {
+    if (actualTarget == QAbstractTexture::Target1DArray ||
+        actualTarget == QAbstractTexture::Target2DArray ||
+        actualTarget == QAbstractTexture::Target2DMultisampleArray ||
+        actualTarget == QAbstractTexture::TargetCubeMapArray) {
         glTex->setLayers(m_properties.layers);
     }
 
-    if (m_actualTarget == QAbstractTexture::Target2DMultisample ||
-        m_actualTarget == QAbstractTexture::Target2DMultisampleArray) {
+    if (actualTarget == QAbstractTexture::Target2DMultisample ||
+        actualTarget == QAbstractTexture::Target2DMultisampleArray) {
         // Set samples count if multisampled texture
         // (multisampled textures don't have mipmaps)
         glTex->setSamples(m_properties.samples);
@@ -471,19 +534,6 @@ QOpenGLTexture *GLTexture::buildGLTexture()
     return glTex;
 }
 
-static void uploadGLData(QOpenGLTexture *glTex,
-                         int level, int layer, QOpenGLTexture::CubeMapFace face,
-                         const QByteArray &bytes, const QTextureImageDataPtr &data)
-{
-    if (data->isCompressed()) {
-        glTex->setCompressedData(level, layer, face, bytes.size(), bytes.constData());
-    } else {
-        QOpenGLPixelTransferOptions uploadOptions;
-        uploadOptions.setAlignment(1);
-        glTex->setData(level, layer, face, data->pixelFormat(), data->pixelType(), bytes.constData(), &uploadOptions);
-    }
-}
-
 void GLTexture::uploadGLTextureData()
 {
     // Upload all QTexImageData set by the QTextureGenerator
@@ -518,23 +568,69 @@ void GLTexture::uploadGLTextureData()
                      static_cast<QOpenGLTexture::CubeMapFace>(m_images[i].face),
                      bytes, imgData);
     }
+    // Free up image data once content has been uploaded
+    // Note: if data functor stores the data, this won't really free anything though
+    m_imageData.clear();
+
+    // Update data from TextureUpdates
+    const QVector<QTextureDataUpdate> textureDataUpdates = std::move(m_pendingTextureDataUpdates);
+    for (const QTextureDataUpdate &update : textureDataUpdates) {
+        const QTextureImageDataPtr imgData = update.data();
+
+        if (!imgData) {
+            qWarning() << Q_FUNC_INFO << "QTextureDataUpdate no QTextureImageData set";
+            continue;
+        }
+
+        const int xOffset = update.x();
+        const int yOffset = update.y();
+        const int zOffset = update.z();
+        const int xExtent = xOffset + imgData->width();
+        const int yExtent = yOffset + imgData->height();
+        const int zExtent = zOffset + imgData->depth();
+
+        // Check update is compatible with our texture
+        if (xOffset >= m_gl->width() ||
+            yOffset >= m_gl->height() ||
+            zOffset >= m_gl->depth() ||
+            xExtent > m_gl->width() ||
+            yExtent > m_gl->height() ||
+            zExtent > m_gl->depth() ||
+            update.mipLevel() >= m_gl->mipLevels() ||
+            update.layer() >= m_gl->layers()) {
+            qWarning() << Q_FUNC_INFO << "QTextureDataUpdate incompatible with texture";
+            continue;
+        }
+
+        const QByteArray bytes = (QTextureImageDataPrivate::get(imgData.get())->m_data);
+        // Here the bytes in the QTextureImageData contain data for a single
+        // layer, face or mip level, unlike the QTextureGenerator case where
+        // they are in a single blob. Hence QTextureImageData::data() is not suitable.
+
+        uploadGLData(m_gl,
+                     update.mipLevel(), update.layer(),
+                     static_cast<QOpenGLTexture::CubeMapFace>(update.face()),
+                     xOffset, yOffset, zOffset,
+                     bytes, imgData);
+    }
 }
 
 void GLTexture::updateGLTextureParameters()
 {
-    const bool isMultisampledTexture = (m_actualTarget == QAbstractTexture::Target2DMultisample ||
-                                        m_actualTarget == QAbstractTexture::Target2DMultisampleArray);
+    const QAbstractTexture::Target actualTarget = m_properties.target;
+    const bool isMultisampledTexture = (actualTarget == QAbstractTexture::Target2DMultisample ||
+                                        actualTarget == QAbstractTexture::Target2DMultisampleArray);
     // Multisampled textures can only be accessed by texelFetch in shaders
     // and don't support wrap modes and mig/mag filtes
     if (isMultisampledTexture)
         return;
 
     m_gl->setWrapMode(QOpenGLTexture::DirectionS, static_cast<QOpenGLTexture::WrapMode>(m_parameters.wrapModeX));
-    if (m_actualTarget != QAbstractTexture::Target1D &&
-        m_actualTarget != QAbstractTexture::Target1DArray &&
-        m_actualTarget != QAbstractTexture::TargetBuffer)
+    if (actualTarget != QAbstractTexture::Target1D &&
+        actualTarget != QAbstractTexture::Target1DArray &&
+        actualTarget != QAbstractTexture::TargetBuffer)
         m_gl->setWrapMode(QOpenGLTexture::DirectionT, static_cast<QOpenGLTexture::WrapMode>(m_parameters.wrapModeY));
-    if (m_actualTarget == QAbstractTexture::Target3D)
+    if (actualTarget == QAbstractTexture::Target3D)
         m_gl->setWrapMode(QOpenGLTexture::DirectionR, static_cast<QOpenGLTexture::WrapMode>(m_parameters.wrapModeZ));
     m_gl->setMinMagFilters(static_cast<QOpenGLTexture::Filter>(m_parameters.minificationFilter),
                            static_cast<QOpenGLTexture::Filter>(m_parameters.magnificationFilter));
@@ -581,8 +677,8 @@ void GLTexture::introspectPropertiesFromSharedTextureId()
     const QPair<int, int> ctxGLVersion = ctx->format().version();
     if (ctxGLVersion.first > 4 || (ctxGLVersion.first == 4 && ctxGLVersion.second >= 5)) {
         // Only for GL 4.5+
-        QOpenGLFunctions_4_5_Core *gl5 = ctx->versionFunctions<QOpenGLFunctions_4_5_Core>();
 #ifdef GL_TEXTURE_TARGET
+        QOpenGLFunctions_4_5_Core *gl5 = ctx->versionFunctions<QOpenGLFunctions_4_5_Core>();
         if (gl5 != nullptr)
             gl5->glGetTextureParameteriv(m_sharedTextureId, GL_TEXTURE_TARGET, reinterpret_cast<int *>(&m_properties.target));
 #endif
